@@ -72,6 +72,12 @@ typedef struct __TypeContext
 
 #define GET_TC(__ptrtc) ((TypeContext *)((__ptrtc)->prv))
 
+// If newObj is set, we should use it rather than JSOBJ
+#define GET_OBJ(__jsobj, __ptrtc) (GET_TC(__ptrtc)->newObj ? GET_TC(__ptrtc)->newObj : __jsobj)
+
+// Avoid infinite loop caused by the default function
+#define DEFAULT_FN_MAX_DEPTH 3
+
 struct PyDictIterState
 {
   PyObject *keys;
@@ -82,22 +88,6 @@ struct PyDictIterState
 //#define PRINTMARK() fprintf(stderr, "%s: MARK(%d)\n", __FILE__, __LINE__)
 #define PRINTMARK()
 
-#ifdef _LP64
-static void *PyIntToINT64(JSOBJ _obj, JSONTypeContext *tc, void *outValue, size_t *_outLen)
-{
-  PyObject *obj = (PyObject *) _obj;
-  *((JSINT64 *) outValue) = PyLong_AsLong (obj);
-  return NULL;
-}
-#else
-static void *PyIntToINT32(JSOBJ _obj, JSONTypeContext *tc, void *outValue, size_t *_outLen)
-{
-  PyObject *obj = (PyObject *) _obj;
-  *((JSINT32 *) outValue) = PyLong_AsLong (obj);
-  return NULL;
-}
-#endif
-
 static void *PyLongToINT64(JSOBJ _obj, JSONTypeContext *tc, void *outValue, size_t *_outLen)
 {
   *((JSINT64 *) outValue) = GET_TC(tc)->longValue;
@@ -432,7 +422,8 @@ static void SetupDictIter(PyObject *dictObj, TypeContext *pc, JSONObjectEncoder
 
 static void Object_beginTypeContext (JSOBJ _obj, JSONTypeContext *tc, JSONObjectEncoder *enc)
 {
-  PyObject *obj, *objRepr, *exc;
+  PyObject *obj, *objRepr, *exc, *defaultFn, *newObj;
+  int level = 0;
   TypeContext *pc;
   PRINTMARK();
   if (!_obj)
@@ -442,6 +433,7 @@ static void Object_beginTypeContext (JSOBJ _obj, JSONTypeContext *tc, JSONObject
   }
 
   obj = (PyObject*) _obj;
+  defaultFn = (PyObject*) enc->prv;
 
   tc->prv = PyObject_Malloc(sizeof(TypeContext));
   pc = (TypeContext *) tc->prv;
@@ -462,6 +454,7 @@ static void Object_beginTypeContext (JSOBJ _obj, JSONTypeContext *tc, JSONObject
   pc->longValue = 0;
   pc->rawJSONValue = NULL;
 
+BEGIN:
   if (PyIter_Check(obj))
   {
     PRINTMARK();
@@ -496,24 +489,14 @@ static void Object_beginTypeContext (JSOBJ _obj, JSONTypeContext *tc, JSONObject
       GET_TC(tc)->unsignedLongValue = PyLong_AsUnsignedLongLong(obj);
 
       exc = PyErr_Occurred();
-      if (exc && PyErr_ExceptionMatches(PyExc_OverflowError))
-      {
-        PRINTMARK();
-        goto INVALID;
-      }
     }
 
-    return;
-  }
-  else
-  if (PyLong_Check(obj))
-  {
-    PRINTMARK();
-#ifdef _LP64
-    pc->PyTypeToJSON = PyIntToINT64; tc->type = JT_LONG;
-#else
-    pc->PyTypeToJSON = PyIntToINT32; tc->type = JT_INT;
-#endif
+    if (exc)
+    {
+      PRINTMARK();
+      goto INVALID;
+    }
+
     return;
   }
   else
@@ -553,7 +536,6 @@ static void Object_beginTypeContext (JSOBJ _obj, JSONTypeContext *tc, JSONObject
     return;
   }
 
-
 ISITERABLE:
   if (PyDict_Check(obj))
   {
@@ -651,6 +633,31 @@ ISITERABLE:
     return;
   }
 
+DEFAULT:
+  if (defaultFn)
+  {
+    // Break infinite loop
+    if (level >= DEFAULT_FN_MAX_DEPTH)
+    {
+      PRINTMARK();
+      PyErr_Format(PyExc_TypeError, "maximum recursion depth exceeded");
+      goto INVALID;
+    }
+
+    newObj = PyObject_CallFunctionObjArgs(defaultFn, obj, NULL);
+    if (newObj)
+    {
+      PRINTMARK();
+      obj = pc->newObj = newObj;
+      level += 1;
+      goto BEGIN;
+    }
+    else
+    {
+      goto INVALID;
+    }
+  }
+
   PRINTMARK();
   PyErr_Clear();
 
@@ -682,12 +689,14 @@ static void Object_endTypeContext(JSOBJ obj, JSONTypeContext *tc)
 
 static const char *Object_getStringValue(JSOBJ obj, JSONTypeContext *tc, size_t *_outLen)
 {
+  obj = GET_OBJ(obj, tc);
   return GET_TC(tc)->PyTypeToJSON (obj, tc, NULL, _outLen);
 }
 
 static JSINT64 Object_getLongValue(JSOBJ obj, JSONTypeContext *tc)
 {
   JSINT64 ret;
+  obj = GET_OBJ(obj, tc);
   GET_TC(tc)->PyTypeToJSON (obj, tc, &ret, NULL);
   return ret;
 }
@@ -695,13 +704,7 @@ static JSINT64 Object_getLongValue(JSOBJ obj, JSONTypeContext *tc)
 static JSUINT64 Object_getUnsignedLongValue(JSOBJ obj, JSONTypeContext *tc)
 {
   JSUINT64 ret;
-  GET_TC(tc)->PyTypeToJSON (obj, tc, &ret, NULL);
-  return ret;
-}
-
-static JSINT32 Object_getIntValue(JSOBJ obj, JSONTypeContext *tc)
-{
-  JSINT32 ret;
+  obj = GET_OBJ(obj, tc);
   GET_TC(tc)->PyTypeToJSON (obj, tc, &ret, NULL);
   return ret;
 }
@@ -709,6 +712,7 @@ static JSINT32 Object_getIntValue(JSOBJ obj, JSONTypeContext *tc)
 static double Object_getDoubleValue(JSOBJ obj, JSONTypeContext *tc)
 {
   double ret;
+  obj = GET_OBJ(obj, tc);
   GET_TC(tc)->PyTypeToJSON (obj, tc, &ret, NULL);
   return ret;
 }
@@ -720,27 +724,31 @@ static void Object_releaseObject(JSOBJ _obj)
 
 static int Object_iterNext(JSOBJ obj, JSONTypeContext *tc)
 {
+  obj = GET_OBJ(obj, tc);
   return GET_TC(tc)->iterNext(obj, tc);
 }
 
 static void Object_iterEnd(JSOBJ obj, JSONTypeContext *tc)
 {
+  obj = GET_OBJ(obj, tc);
   GET_TC(tc)->iterEnd(obj, tc);
 }
 
 static JSOBJ Object_iterGetValue(JSOBJ obj, JSONTypeContext *tc)
 {
+  obj = GET_OBJ(obj, tc);
   return GET_TC(tc)->iterGetValue(obj, tc);
 }
 
 static char *Object_iterGetName(JSOBJ obj, JSONTypeContext *tc, size_t *outLen)
 {
+  obj = GET_OBJ(obj, tc);
   return GET_TC(tc)->iterGetName(obj, tc, outLen);
 }
 
 PyObject* objToJSON(PyObject* self, PyObject *args, PyObject *kwargs)
 {
-  static char *kwlist[] = { "obj", "ensure_ascii", "encode_html_chars", "escape_forward_slashes", "sort_keys", "indent", "allow_nan", "reject_bytes", NULL };
+  static char *kwlist[] = { "obj", "ensure_ascii", "encode_html_chars", "escape_forward_slashes", "sort_keys", "indent", "allow_nan", "reject_bytes", "default", NULL };
 
   char buffer[65536];
   char *ret;
@@ -751,6 +759,7 @@ PyObject* objToJSON(PyObject* self, PyObject *args, PyObject *kwargs)
   PyObject *oencodeHTMLChars = NULL;
   PyObject *oescapeForwardSlashes = NULL;
   PyObject *osortKeys = NULL;
+  PyObject *odefaultFn = NULL;
   int allowNan = -1;
   int orejectBytes = -1;
 
@@ -761,7 +770,6 @@ PyObject* objToJSON(PyObject* self, PyObject *args, PyObject *kwargs)
     Object_getStringValue,
     Object_getLongValue,
     Object_getUnsignedLongValue,
-    Object_getIntValue,
     Object_getDoubleValue,
     Object_iterNext,
     Object_iterEnd,
@@ -785,7 +793,7 @@ PyObject* objToJSON(PyObject* self, PyObject *args, PyObject *kwargs)
 
   PRINTMARK();
 
-  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOOOiii", kwlist, &oinput, &oensureAscii, &oencodeHTMLChars, &oescapeForwardSlashes, &osortKeys, &encoder.indent, &allowNan, &orejectBytes))
+  if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OOOOiiiO", kwlist, &oinput, &oensureAscii, &oencodeHTMLChars, &oescapeForwardSlashes, &osortKeys, &encoder.indent, &allowNan, &orejectBytes, &odefaultFn))
   {
     return NULL;
   }
@@ -815,6 +823,12 @@ PyObject* objToJSON(PyObject* self, PyObject *args, PyObject *kwargs)
     encoder.allowNan = allowNan;
   }
 
+  if (odefaultFn != NULL && odefaultFn != Py_None)
+  {
+    // Here use prv to store default function
+    encoder.prv = odefaultFn;
+  }
+
   if (encoder.allowNan)
   {
     csInf = "Inf";
