@@ -1,10 +1,13 @@
+import datetime as dt
 import decimal
 import io
 import json
 import math
 import re
 import sys
+import uuid
 from collections import OrderedDict
+from pathlib import Path
 
 import pytest
 import ujson
@@ -78,6 +81,12 @@ def test_double_long_decimal_issue():
     assert sut == decoded
 
 
+# NOTE: can't match exponents -9 to -5; Python 0-pads
+@pytest.mark.parametrize("val", [1e-10, 1e-4, 1e10, 1e15, 1e16, 1e30])
+def test_encode_float_string_rep(val):
+    assert ujson.dumps(val) == json.dumps(val)
+
+
 def test_encode_decode_long_decimal():
     sut = {"a": -528656961.4399388}
     encoded = ujson.dumps(sut)
@@ -249,11 +258,21 @@ def test_encode_to_utf8():
     assert dec == json.loads(enc)
 
 
-def test_encode_indent():
-    test_input = '{\n    "obj": 31337\n}'
+@pytest.mark.parametrize(
+    "test_input",
+    [
+        '{\n    "obj": 31337\n}',
+        "{}",
+        "[]",
+        '{\n    "a": {}\n}',
+        "[\n    []\n]",
+    ],
+)
+def test_encode_indent(test_input):
     obj = ujson.decode(test_input)
     output = ujson.encode(obj, indent=4)
     assert test_input == output
+    assert output == json.dumps(obj, indent=4)
 
 
 def test_decode_from_unicode():
@@ -378,11 +397,11 @@ def test_decode_number_with32bit_sign_bit():
     # sign bit set (2**31 <= x < 2**32) are decoded properly.
     docs = (
         '{"id": 3590016419}',
-        '{"id": %s}' % 2 ** 31,
-        '{"id": %s}' % 2 ** 32,
-        '{"id": %s}' % ((2 ** 32) - 1),
+        '{"id": %s}' % 2**31,
+        '{"id": %s}' % 2**32,
+        '{"id": %s}' % ((2**32) - 1),
     )
-    results = (3590016419, 2 ** 31, 2 ** 32, 2 ** 32 - 1)
+    results = (3590016419, 2**31, 2**32, 2**32 - 1)
     for doc, result in zip(docs, results):
         assert ujson.decode(doc)["id"] == result
 
@@ -509,6 +528,7 @@ def test_decode_no_assert(test_input):
     [
         ("31337", 31337),
         ("-31337", -31337),
+        ("100000000000000000000.0", 1e20),
     ],
 )
 def test_decode(test_input, expected):
@@ -536,46 +556,46 @@ def test_decode_numeric_int_exp(test_input):
 @pytest.mark.parametrize(
     "test_input, expected",
     [
-        ('{{1337:""}}', ValueError),  # broken dict key type leak test
-        ('{{"key":"}', ValueError),  # broken dict leak test
-        ('{{"key":"}', ValueError),  # broken dict leak test
-        ("[[[true", ValueError),  # broken list leak test
+        ('{{1337:""}}', ujson.JSONDecodeError),  # broken dict key type leak test
+        ('{{"key":"}', ujson.JSONDecodeError),  # broken dict leak test
+        ('{{"key":"}', ujson.JSONDecodeError),  # broken dict leak test
+        ("[[[true", ujson.JSONDecodeError),  # broken list leak test
     ],
 )
 def test_decode_range_raises(test_input, expected):
     for x in range(1000):
-        with pytest.raises(ValueError):
+        with pytest.raises(expected):
             ujson.decode(test_input)
 
 
 @pytest.mark.parametrize(
     "test_input, expected",
     [
-        ("fdsa sda v9sa fdsa", ValueError),  # jibberish
-        ("[", ValueError),  # broken array start
-        ("{", ValueError),  # broken object start
-        ("]", ValueError),  # broken array end
-        ("}", ValueError),  # broken object end
-        ('{"one":1,}', ValueError),  # object trailing comma fail
-        ('"TESTING', ValueError),  # string unterminated
-        ('"TESTING\\"', ValueError),  # string bad escape
-        ("tru", ValueError),  # true broken
-        ("fa", ValueError),  # false broken
-        ("n", ValueError),  # null broken
-        ("{{{{31337}}}}", ValueError),  # dict with no key
-        ('{{{{"key"}}}}', ValueError),  # dict with no colon or value
-        ('{{{{"key":}}}}', ValueError),  # dict with no value
-        ("[31337,]", ValueError),  # array trailing comma fail
-        ("[,31337]", ValueError),  # array leading comma fail
-        ("[,]", ValueError),  # array only comma fail
-        ("[]]", ValueError),  # array unmatched bracket fail
-        ("18446744073709551616", ValueError),  # too big value
-        ("-90223372036854775809", ValueError),  # too small value
-        ("18446744073709551616", ValueError),  # very too big value
-        ("-90223372036854775809", ValueError),  # very too small value
-        ("{}\n\t a", ValueError),  # with trailing non whitespaces
-        ("[18446744073709551616]", ValueError),  # array with big int
-        ('{"age", 44}', ValueError),  # read bad object syntax
+        ("fdsa sda v9sa fdsa", ujson.JSONDecodeError),  # jibberish
+        ("[", ujson.JSONDecodeError),  # broken array start
+        ("{", ujson.JSONDecodeError),  # broken object start
+        ("]", ujson.JSONDecodeError),  # broken array end
+        ("}", ujson.JSONDecodeError),  # broken object end
+        ('{"one":1,}', ujson.JSONDecodeError),  # object trailing comma fail
+        ('"TESTING', ujson.JSONDecodeError),  # string unterminated
+        ('"TESTING\\"', ujson.JSONDecodeError),  # string bad escape
+        ("tru", ujson.JSONDecodeError),  # true broken
+        ("fa", ujson.JSONDecodeError),  # false broken
+        ("n", ujson.JSONDecodeError),  # null broken
+        ("{{{{31337}}}}", ujson.JSONDecodeError),  # dict with no key
+        ('{{{{"key"}}}}', ujson.JSONDecodeError),  # dict with no colon or value
+        ('{{{{"key":}}}}', ujson.JSONDecodeError),  # dict with no value
+        ("[31337,]", ujson.JSONDecodeError),  # array trailing comma fail
+        ("[,31337]", ujson.JSONDecodeError),  # array leading comma fail
+        ("[,]", ujson.JSONDecodeError),  # array only comma fail
+        ("[]]", ujson.JSONDecodeError),  # array unmatched bracket fail
+        ("18446744073709551616", ujson.JSONDecodeError),  # too big value
+        ("-90223372036854775809", ujson.JSONDecodeError),  # too small value
+        ("18446744073709551616", ujson.JSONDecodeError),  # very too big value
+        ("-90223372036854775809", ujson.JSONDecodeError),  # very too small value
+        ("{}\n\t a", ujson.JSONDecodeError),  # with trailing non whitespaces
+        ("[18446744073709551616]", ujson.JSONDecodeError),  # array with big int
+        ('{"age", 44}', ujson.JSONDecodeError),  # read bad object syntax
     ],
 )
 def test_decode_raises(test_input, expected):
@@ -586,8 +606,8 @@ def test_decode_raises(test_input, expected):
 @pytest.mark.parametrize(
     "test_input, expected",
     [
-        ("[", ValueError),  # array depth too big
-        ("{", ValueError),  # object depth too big
+        ("[", ujson.JSONDecodeError),  # array depth too big
+        ("{", ujson.JSONDecodeError),  # object depth too big
     ],
 )
 def test_decode_raises_for_long_input(test_input, expected):
@@ -595,6 +615,11 @@ def test_decode_raises_for_long_input(test_input, expected):
         ujson.decode(test_input * (1024 * 1024))
 
 
+def test_decode_exception_is_value_error():
+    assert issubclass(ujson.JSONDecodeError, ValueError)
+    assert ujson.JSONDecodeError is not ValueError
+
+
 @pytest.mark.parametrize(
     "test_input, expected",
     [
@@ -635,6 +660,10 @@ def test_dumps_raises(test_input, expected_exception, expected_message):
         (float("inf"), OverflowError),
         (-float("inf"), OverflowError),
         (12839128391289382193812939, OverflowError),
+        ([12839128391289382193812939], OverflowError),
+        ([12839128391289382193812939, 42], OverflowError),
+        ({"a": 12839128391289382193812939}, OverflowError),
+        ({"a": 12839128391289382193812939, "b": 42}, OverflowError),
     ],
 )
 def test_encode_raises_allow_nan(test_input, expected_exception):
@@ -642,6 +671,79 @@ def test_encode_raises_allow_nan(test_input, expected_exception):
         ujson.dumps(test_input, allow_nan=False)
 
 
+def test_nan_inf_support():
+    # Test ported from pandas
+    text = '["a", NaN, "NaN", Infinity, "Infinity", -Infinity, "-Infinity"]'
+    data = ujson.loads(text)
+    expected = [
+        "a",
+        float("nan"),
+        "NaN",
+        float("inf"),
+        "Infinity",
+        -float("inf"),
+        "-Infinity",
+    ]
+    for a, b in zip(data, expected):
+        assert a == b or math.isnan(a) and math.isnan(b)
+
+
+def test_special_singletons():
+    pos_inf = ujson.loads("Infinity")
+    neg_inf = ujson.loads("-Infinity")
+    nan = ujson.loads("NaN")
+    null = ujson.loads("null")
+    assert math.isinf(pos_inf) and pos_inf > 0
+    assert math.isinf(neg_inf) and neg_inf < 0
+    assert math.isnan(nan)
+    assert null is None
+
+
+@pytest.mark.parametrize(
+    "test_input, expected_message",
+    [
+        ("n", "Unexpected character .* 'null'"),
+        ("N", "Unexpected character .*'NaN'"),
+        ("NA", "Unexpected character .* 'NaN'"),
+        ("Na N", "Unexpected character .* 'NaN'"),
+        ("nan", "Unexpected character .* 'null'"),
+        ("none", "Unexpected character .* 'null'"),
+        ("i", "Expected object or value"),
+        ("I", "Unexpected character .* 'Infinity'"),
+        ("Inf", "Unexpected character .* 'Infinity'"),
+        ("InfinitY", "Unexpected character .* 'Infinity'"),
+        ("-i", "Trailing data"),
+        ("-I", "Unexpected character .* '-Infinity'"),
+        ("-Inf", "Unexpected character .* '-Infinity'"),
+        ("-InfinitY", "Unexpected character .* '-Infinity'"),
+        ("- i", "Trailing data"),
+        ("- I", "Trailing data"),
+        ("- Inf", "Trailing data"),
+        ("- InfinitY", "Trailing data"),
+    ],
+)
+def test_incomplete_special_inputs(test_input, expected_message):
+    with pytest.raises(ujson.JSONDecodeError, match=expected_message):
+        print(f"test_input = {test_input!r}")
+        ujson.loads(test_input)
+
+
+@pytest.mark.parametrize(
+    "test_input, expected_message",
+    [
+        ("NaNaNaN", "Trailing data"),
+        ("Infinity and Beyond", "Trailing data"),
+        ("-Infinity-and-Beyond", "Trailing data"),
+        ("NaN!", "Trailing data"),
+        ("Infinity!", "Trailing data"),
+        ("-Infinity!", "Trailing data"),
+    ],
+)
+def test_overcomplete_special_inputs(test_input, expected_message):
+    with pytest.raises(ujson.JSONDecodeError, match=expected_message):
+        ujson.loads(test_input)
+
+
 @pytest.mark.parametrize(
     "test_input",
     [
@@ -806,6 +908,72 @@ def test_reject_bytes_false():
     assert ujson.dumps(data, reject_bytes=False) == '{"a":"b"}'
 
 
+def test_encode_none_key():
+    data = {None: None}
+    assert ujson.dumps(data) == '{"null":null}'
+
+
+def test_default_function():
+    iso8601_time_format = "%Y-%m-%dT%H:%M:%S.%f"
+
+    class CustomObject:
+        pass
+
+    class UnjsonableObject:
+        pass
+
+    def default(value):
+        if isinstance(value, dt.datetime):
+            return value.strftime(iso8601_time_format)
+        elif isinstance(value, uuid.UUID):
+            return value.hex
+        elif isinstance(value, CustomObject):
+            raise ValueError("invalid value")
+        return value
+
+    now = dt.datetime.now()
+    expected_output = '"%s"' % now.strftime(iso8601_time_format)
+    assert ujson.dumps(now, default=default) == expected_output
+
+    uuid4 = uuid.uuid4()
+    expected_output = '"%s"' % uuid4.hex
+    assert ujson.dumps(uuid4, default=default) == expected_output
+
+    custom_obj = CustomObject()
+    with pytest.raises(ValueError, match="invalid value"):
+        ujson.dumps(custom_obj, default=default)
+
+    unjsonable_obj = UnjsonableObject()
+    with pytest.raises(TypeError, match="maximum recursion depth exceeded"):
+        ujson.dumps(unjsonable_obj, default=default)
+
+
+@pytest.mark.parametrize("indent", list(range(65537, 65542)))
+def test_dump_huge_indent(indent):
+    ujson.encode({"a": True}, indent=indent)
+
+
+@pytest.mark.parametrize("first_length", list(range(2, 7)))
+@pytest.mark.parametrize("second_length", list(range(10919, 10924)))
+def test_dump_long_string(first_length, second_length):
+    ujson.dumps(["a" * first_length, "\x00" * second_length])
+
+
+def test_dump_indented_nested_list():
+    a = _a = []
+    for i in range(20):
+        _a.append(list(range(i)))
+        _a = _a[-1]
+        ujson.dumps(a, indent=i)
+
+
+@pytest.mark.parametrize("indent", [0, 1, 2, 4, 5, 8, 49])
+def test_issue_334(indent):
+    path = Path(__file__).with_name("334-reproducer.json")
+    a = ujson.loads(path.read_bytes())
+    ujson.dumps(a, indent=indent)
+
+
 """
 def test_decode_numeric_int_frc_overflow():
 input = "X.Y"
